// ------------------------------ mathfun.c ------------------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2025 Pascal Baillehache baillehache.pascal@gmail.com
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#include "mathfun.h"
#include "capymath.h"

// 'that' keyword
#define CapyThatMathFun CapyMathFun* that = (CapyMathFun*)capyThat
#define CapyThatHyperplane CapyHyperplane* that = (CapyHyperplane*)capyThat

// Delta used to evaluate the derivative of the function
#define CAPY_DERIVATIVE_EPSILON 1e-9
#define CAPY_DERIVATIVE_INV_EPSILON 1e9

// Evaluate the function for a given input.
// Input:
//    in: the input vector
//   out: array of double updated with the result of evaluation
static void Eval(
  double const* const in,
        double* const out) {
  (void)in; (void)out;
  raiseExc(CapyExc_UndefinedExecution);
  assert(false && "MathFun.eval is undefined.");
}

// Evaluate the derivative of the function for a given input and
// dimension
// Input:
//     in: the input vector
//   iDim: the derivation dimension
//    out: array of dimOut double updated with the result of derivation
// Exceptions:
//   May raise CapyExc_MallocFailed
static void EvalDerivative(
  double* const in,
   size_t const iDim,
  double* const out) {
  CapyThatMathFun;
  double const backupVal = in[iDim];
  in[iDim] = in[iDim] + 0.5 * CAPY_DERIVATIVE_EPSILON;
  $(that, eval)(in, out);
  in[iDim] -= CAPY_DERIVATIVE_EPSILON;
  $(that, eval)(in, that->outBuffer);
  loop(iOut, that->dimOut) {
    out[iOut] -= that->outBuffer[iOut];
    out[iOut] *= CAPY_DERIVATIVE_INV_EPSILON;
  }
  in[iDim] = backupVal;
}

// Recursive function for the evaluation of the integration
static void EvalIntegralRec(
            CapyMathFun* const that,
                 double* const in,
  CapyRangeDouble const* const domains,
                 double* const out,
                  size_t const iDim) {
  if(iDim < that->dimIn - 1) {
    double u[2][that->dimOut];
    loop(i, 2) {
      in[iDim] = domains[iDim].vals[i];
      EvalIntegralRec(that, in, domains, u[i], iDim + 1);
    }
    loop(i, that->dimOut) {
      out[i] =
        0.5 * (domains[iDim].max - domains[iDim].min) * (u[0][i] + u[1][i]);
    }
  } else {
    double u[2][that->dimOut];
    loop(i, 2) {
      in[iDim] = domains[iDim].vals[i];
      $(that, eval)(in, u[i]);
    }
    loop(i, that->dimOut) {
      out[i] =
        0.5 * (domains[iDim].max - domains[iDim].min) * (u[0][i] + u[1][i]);
    }
  }
}

// Evaluate the integration of the function over a domain
// Input:
//   domains: the domain over which to integrate, array of that->dimIn ranges
//   out: the result of integration
// Output:
//   'out' is updated. Uses the trapezoidal rule to approximate the integration.
static void EvalIntegral(
  CapyRangeDouble const* const domains,
                 double* const out) {
  CapyThatMathFun;
  double* in = NULL;
  safeMalloc(in, that->dimIn);
  if(in == NULL) return;
  EvalIntegralRec(that, in, domains, out, 0);
  free(in);
}

// Evaluate the jacobian (aka gradient if dimOut==1) of the function for a given
// input
// Input:
//          in: the input vector
//    jacobian: array of dimIn*dimOut double updated with the jacobian,
//              column i is derivative in dimension i, stored by row
// Exceptions:
//   May raise CapyExc_MallocFailed
static void EvalJacobian(
  double* const in,
  double* const jacobian) {
  CapyThatMathFun;
  CapyMat jacobianTransp = {.vals = NULL};
  try {
    jacobianTransp = CapyMatCreate(that->dimOut, that->dimIn);
    loop(i, that->dimIn) {
      $(that, evalDerivative)(
        in, i, jacobianTransp.vals + i * jacobianTransp.nbCol);
    }
    CapyMat j = {
      .nbCol = jacobianTransp.nbRow,
      .nbRow = jacobianTransp.nbCol,
      .vals = jacobian
    };
    CapyMatTransp(&jacobianTransp, &j);
  } endCatch;
  CapyMatDestruct(&jacobianTransp);
  CapyForwardExc();
}

// Evaluate the divergence of the function for a given input
// Input:
//     in: the input vector
// Output:
//   Return the divergence.
// Exceptions:
//   May raise CapyExc_MallocFailed
static double EvalDivergence(double* const in) {
  CapyThatMathFun;
  struct {
    double* out;
    double divergence;
  } volatiles = {NULL, 0.0};
  try {
    safeMalloc(volatiles.out, that->dimOut);
    if(!(volatiles.out)) return 0.0;
    loop(iDim, that->dimIn) {
      $(that, evalDerivative)(in, iDim, volatiles.out);
      volatiles.divergence += volatiles.out[iDim];
    }
  } endCatch;
  free(volatiles.out);
  CapyForwardExc();
  return volatiles.divergence;
}

// Find the solution of f(in)=out using that->nbIterSolveNewtonMethod (by
// default: 10) iterations of the Newton method
// Input:
//    in: initial approximation of the solution, updated with the result
//   out: output values
// Exceptions:
//   May raise CapyExc_MallocFailed
static void SolveByNewtonMethod(
        double* const in,
  double const* const out) {
  CapyThatMathFun;

  // cf https://en.wikipedia.org/wiki/Newton%27s_method#Generalizations
  CapyVec out2 = { .dim = 0, .vals = NULL };
  CapyMat jacobian = { .nbCol = 0, .nbRow = 0, .vals = NULL };
  CapyMat invJacobian = { .nbCol = 0, .nbRow = 0, .vals = NULL };
  try {
    out2 = CapyVecCreate(that->dimOut);
    jacobian = CapyMatCreate(that->dimIn, that->dimOut);
    invJacobian = CapyMatCreate(that->dimOut, that->dimIn);
    loop(iIter, that->nbIterSolveNewtonMethod) {
      $(that, eval)(in, out2.vals);
      loop(i, that->dimOut) out2.vals[i] -= out[i];
      $(that, evalJacobian)(in, jacobian.vals);
      CapyMatInv(&jacobian, &invJacobian);
      loop(i, that->dimIn) loop(j , that->dimOut) {
        in[i] -= invJacobian.vals[i * that->dimOut + j] * out2.vals[j];
      }
    }
  } endCatch;
  CapyVecDestruct(&out2);
  CapyMatDestruct(&jacobian);
  CapyMatDestruct(&invJacobian);
  CapyForwardExc();
}

// Free the memory used by a CapyMathFun
static void Destruct(void) {
  CapyThatMathFun;
  free(that->outBuffer);
  free(that->domains);
}

// Create a CapyMathFun
// Input:
//    dimIn: input dimension
//   dimOut: output dimension
// Output:
//   Return a CapyMathFun
CapyMathFun CapyMathFunCreate(
  size_t const dimIn,
  size_t const dimOut) {
  CapyMathFun that = {
    .dimIn = dimIn,
    .dimOut = dimOut,
    .domains = NULL,
    .nbIterSolveNewtonMethod = 10,
    .destruct = Destruct,
    .eval = Eval,
    .evalDerivative = EvalDerivative,
    .solveByNewtonMethod = SolveByNewtonMethod,
    .evalJacobian = EvalJacobian,
    .evalDivergence = EvalDivergence,
    .evalIntegral = EvalIntegral,
  };
  safeMalloc(that.outBuffer, dimOut);
  safeMalloc(that.domains, dimIn);
  if(that.domains) loop(iIn, dimIn) {
    that.domains[iIn] = CapyRangeDoubleCreate(0.0, 1.0);
  }
  return that;
}

// Evaluate the hyperplane function for a given input.
// Input:
//    in: the input vector
//   out: array of double updated with the result of evaluation
static void EvalHyperplane(
  double const* const in,
        double* const out) {
  CapyThatHyperplane;
  out[0] = 0.0;
  loop(i, that->dimIn) out[0] += in[i] * that->u.vals[i];
  out[0] += that->u.vals[that->dimIn];
}

// Free the memory used by a CapyHyperplane
// Input:
//   that: the CapyHyperplane to free
static void HyperplaneDestruct(void) {
  CapyThatHyperplane;
  $(that, destructCapyMathFun)();
  CapyVecDestruct(&(that->u));
}

// Create a CapyHyperplane
// Input:
//    dimIn: input dimension
// Output:
//   Return a CapyHyperplane.
CapyHyperplane CapyHyperplaneCreate(size_t const dimIn) {
  CapyHyperplane that;
  CapyInherits(that, CapyMathFun, (dimIn, 1));
  that.u = CapyVecCreate(dimIn + 1);
  that.eval = EvalHyperplane;
  that.destruct = HyperplaneDestruct;
  return that;
}

// Allocate memory and create a CapyHyperplane
// Input:
//    dimIn: input dimension
// Output:
//   Return a CapyHyperplane.
CapyHyperplane* CapyHyperplaneAlloc(size_t const dimIn) {
  CapyHyperplane* that = NULL;
  safeMalloc(that, 1);
  if(!that) return NULL;
  *that = CapyHyperplaneCreate(dimIn);
  return that;
}

// Free the memory used by a CapyHyperplane* and reset '*that' to
// NULL
// Input:
//   that: the CapyHyperplane to free
void CapyHyperplaneFree(CapyHyperplane** const that) {
  if(that == NULL || *that == NULL) return;
  $(*that, destruct)();
  free(*that);
  *that = NULL;
}
