// ------------------------------- mathfun.h -------------------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2025 Pascal Baillehache baillehache.pascal@gmail.com
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CAPY_MATHFUN_H
#define CAPY_MATHFUN_H
#include "externalHeaders.h"
#include "cext.h"
#include "range.h"
#include "capymath.h"

// Description:
// Class to manipulate mathematical functions.

// CapyMathFun object definition macro
#define CapyMathFunDef {                   \
  union {                                  \
    size_t dims[2];                        \
    struct __attribute__((packed)) {       \
      size_t dimIn, dimOut;                \
    };                                     \
  };                                       \
  size_t nbIterSolveNewtonMethod;          \
  CapyRangeDouble* domains;                \
  double* outBuffer;                       \
  void (*destruct)(void);                  \
  void (*eval)(                            \
    double const* const in,                \
          double* const out);              \
  void (*evalJacobian)(                    \
    double* const in,                      \
    double* const jacobian);               \
  void (*evalDerivative)(                  \
       double* const in,                   \
        size_t const iDim,                 \
       double* const out);                 \
  void (*evalIntegral)(                    \
    CapyRangeDouble const* const domains,  \
                   double* const out);     \
  double (*evalDivergence)(                \
    double* const in);                     \
  void (*solveByNewtonMethod)(             \
          double* const in,                \
    double const* const out);              \
}

// CapyMathFun object
typedef struct CapyMathFun CapyMathFunDef CapyMathFun;

// Create a CapyMathFun
// Input:
//    dimIn: input dimension
//   dimOut: output dimension
// Output:
//   Return a CapyMathFun. domains set by default to [0, 1]
CapyMathFun CapyMathFunCreate(
  size_t const dimIn,
  size_t const dimOut);

// Hyperplane function
typedef struct CapyHyperplane {

  // Inherit CapyMathFun
  struct CapyMathFunDef;

  // Hyperplane parameters, the hyperplane is defined as the points X such as
  // H(X)=u.[X|1]=0
  CapyVec u;

  // Destructor
  void (*destructCapyMathFun)(void);
} CapyHyperplane;

// Create a CapyHyperplane
// Input:
//    dimIn: input dimension
// Output:
//   Return a CapyHyperplane.
CapyHyperplane CapyHyperplaneCreate(size_t const dimIn);

// Allocate memory and create a CapyHyperplane
// Input:
//    dimIn: input dimension
// Output:
//   Return a CapyHyperplane.
CapyHyperplane* CapyHyperplaneAlloc(size_t const dimIn);

// Free the memory used by a CapyHyperplane* and reset '*that' to
// NULL
// Input:
//   that: the CapyHyperplane to free
void CapyHyperplaneFree(CapyHyperplane** const that);
#endif
